
/* Web Polygraph       http://www.web-polygraph.org/
 * (C) 2003-2006 The Measurement Factory
 * Licensed under the Apache License, Version 2.0 */

#include "base/polygraph.h"

#include "base/StatIntvlRec.h"
#include "base/polyLogCats.h"
#include "runtime/LogComment.h"
#include "csm/BodyIter.h"
#include "csm/ContentMgr.h"
#include "runtime/globals.h"
#include "runtime/ErrorMgr.h"
#include "runtime/polyErrors.h"
#include "runtime/FtpText.h"
#include "runtime/SharedOpts.h"
#include "server/Server.h"
#include "server/SrvOpts.h"

#include "server/FtpSrvXact.h"

FtpSrvXact::FtpSrvXact(): theDataPort(-1),
    theDataXferState(dxsNone), theDataIoState(disNone) {
	FtpSrvXact::reset();
}

void FtpSrvXact::reset() {
	SrvXact::reset();
	protoStat = &StatIntvlRec::theFtpStat;
	theReqOid.type(TheBodilessContentId);
	theReq.reset();
	theDir = String();
	theFilename = String();
	stopDataListen();
	closeDataChannel();
	theDataXferState = dxsNone;
	theDataConn.reset();
	theDataConn.logCat(lgcSrvSide);
	theWrittenSize.reset();
}

void FtpSrvXact::finish(Error err) {
	stopDataListen();
	closeDataChannel();
	theRepSize = theWrittenSize; // XXX: ugly; what is reply?
	if (!theOid && !err) // we have not seen RETR or errors
		err = errFtpNoDataXfer;
	SrvXact::finish(err);
}

void FtpSrvXact::noteReadReady(int fd) {
	if (fd == theSock.fd())
		acceptDataConnection();
	else
		SrvXact::noteReadReady(fd);
}

void FtpSrvXact::noteWriteReady(int fd) {
	if (fd == theDataConn.fd())
		noteDataChannelWriteReady();
	else
		SrvXact::noteWriteReady(fd);
}

void FtpSrvXact::doStart() {
	newState(stHdrWaiting);
	theWrittenSize.expect(0); // will increase later
	SrvXact::doStart();
}

void FtpSrvXact::noteConnReady() {
	// Connection has just been established, send greeting
	newState(stSpaceWaiting);
	noteWriteReady(theConn->fd());
}

void FtpSrvXact::noteHdrDataReady() {
	IOBuf &buf = theConn->theRdBuf;
	if (theReq.parse(buf.content(), buf.contSize())) {
		theReqSize.expect(theReq.size());
//Comment << "FTP request (" << buf.contSize() << "): " << endc; printMsg(buf.content(), buf.contSize());
		if (const Error err = interpretCmd()) {
			finish(err);
			return;
		}
		// XXX: check for another command and quit, if any. Since we do not
		// and should not format replies before we can write them, we cannot
		// handle pipelined commands. The I/O code will not call us if there
		// are buffered leftovers but no new I/O activity.
		if (theDataXferState != dxsConnWaiting) {
			newState(stSpaceWaiting); // switch to writing ctrl response
			theConn->theRd.stop(this);
		} // else continue reading (but there should be nothing to read)
		consume(theReq.size());
		return;
	}

	if (theConn->theRdBuf.full()) { // command too big
		finish(errFtpHugeCmd);
		return;
	}

	if (theConn->atEof()) {
		if (!theConn->theRdBuf.empty())
			finish(errFtpPrematureEndOfMsg); // we did not get the entire cmd
		else
        if (theDataXferState == dxsInProgress)
			finish(errFtpPrematureEndOfCtrl); // control channel died first
		else
			finish(0); // assume the user decided not to send QUIT
		return;
	}

	// continue reading
}

void FtpSrvXact::noteRepSent() {
	theWrittenSize.got(theRepSize.actual());
	theRepSize.reset();

	if (theReq.cmd() == FtpReq::frcQUIT) {
		theConn->lastUse(true);
		finish(0);
		return;
	}

	// expect more commands
	newState(stHdrWaiting);
	if (!theConn->theRd.theReserv) 
		theConn->theRd.start(this);

	// after response is sent, start sending data if needed 
	if (theDataXferState == dxsInProgress) {
		Assert(theDataIoState == disBusy);
		if (!theDataConn.theWr.theReserv)
			theDataConn.theWr.start(this);
	}
}

void FtpSrvXact::makeRep(WrBuf &buf) {
	Assert(!theRepSize.expected().known());
	ofixedstream os(buf.space(), buf.spaceSize());
	switch(theReq.cmd()) {
		case FtpReq::frcUnknown:
			makeSimpleResponse(os, ftpRspGreeting);
			break;
		case FtpReq::frcSYST:
			makeSimpleResponse(os, ftpRspSystem);
			break;
		case FtpReq::frcUSER:
			makeSimpleResponse(os, ftpRspUser);
			break;
		case FtpReq::frcPASS:
			makeSimpleResponse(os, ftpRspPass);
			break;
		case FtpReq::frcTYPE:
			makeSimpleResponse(os, ftpRspTypeBinary);
			break;
		case FtpReq::frcCWD:
			makeSimpleResponse(os, ftpRspCwd);
			break;
		case FtpReq::frcMDTM:
			makeMdtmResponse(os);
			break;
		case FtpReq::frcSIZE:
			makeSizeResponse(os);
			break;
		case FtpReq::frcPASV:
			makePasvResponse(os);
			break;
		case FtpReq::frcRETR:
			makeRetrResponse(os);
			break;
		case FtpReq::frcQUIT:
			makeSimpleResponse(os, ftpRspQuit);
			break;
		case FtpReq::frcEPSV:
		case FtpReq::frcOther:
			makeSimpleResponse(os, ftpRspCmdNotImplemented);
			break;
		default:
			Assert(false);
	}
	theRepSize.expect(Size(os.tellp()));
	theWrittenSize.expectMore(theRepSize.expected());
	buf.appended(theRepSize.expected());
//Comment << "FTP reply (" << buf.contSize() << "): " << endc; printMsg(buf.content(), buf.contSize());
}

Error FtpSrvXact::interpretCmd() {
	Error err = 0;
	switch(theReq.cmd()) {
		case FtpReq::frcUSER:
		case FtpReq::frcPASS:
		case FtpReq::frcQUIT:
		case FtpReq::frcSYST:
			break;
		case FtpReq::frcTYPE:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
			if (theReq.argument() != "I" &&
				theReq.argument() != "L 8")
				err = errFtpModeNotBinary;
			break;
		case FtpReq::frcCWD:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
			if (theReq.argument() == "/")
				theDir = ""; // root directory
			else
			if (theReq.argument().startsWith("/"))
				theDir = theReq.argument(); // absolute directory
			else
				theDir += "/" + theReq.argument(); // relative
			break;
		case FtpReq::frcMDTM:
		case FtpReq::frcSIZE:
			if (!theReq.argument())
				err = errFtpNoArgument;
			else
				theFilename = theReq.argument();
			break;
		case FtpReq::frcPASV:
			err = startDataListen();
			break;
		case FtpReq::frcRETR:
			err = processRetr();
			break;
		case FtpReq::frcEPSV:
			// we do not support it, but do not want to report it
			// TODO: add something like ReportErrorOnce(err, key)
			break;
		case FtpReq::frcOther:
			// report but continue
			if (ReportError(errFtpCommandNotImplemented) &&
				TheOpts.theDumpFlags(dumpErr, dumpAny))
                printMsg(theReq.image().data(), theReq.image().len());
			break;
		default:
			err = errFtpSyntax;
	}
	return err;
}

Error FtpSrvXact::startDataListen() {
	if (theDataIoState != disNone) {
		if (theDataXferState != dxsNone) // cannot reset connection now
			return errFtpCmdSequence;
		stopDataListen();
		closeDataChannel();
	}
	
	Assert(theDataIoState == disNone);
	Assert(theDataPort < 0);

	// TODO: support port manager so that we can use non-ephemeral ports
	const NetAddr addr(theOwner->host().addrN(), 0);
	theSock = theOwner->makeListenSocket(addr);
	if (!theSock) {
		if (ReportError2(Error::Last(), lgcCltSide))
			Comment << "OS probably ran out of ephemeral ports at " << addr << endc;
		return errOther;
	}

	theDataPort = theSock.lport();
	Assert(theDataPort >= 0);

	theReserv = TheFileScanner->setFD(theSock.fd(), dirRead, this);
	theDataIoState = disListen;
//Comment << theSock.fd() << " listenting on " << addr << endc;
	return 0;
}

void FtpSrvXact::stopDataListen() {
	if (theDataIoState == disListen) {
		if (theReserv)
			TheFileScanner->clearRes(theReserv);
		if (theSock)
			theSock.close();
		theDataPort = -1;
		theDataIoState = disNone;
	}
}

void FtpSrvXact::closeDataChannel() {
	if (theDataIoState == disBusy) {
		theDataConn.closeNow();
		theDataIoState = disNone;
	}
}

void FtpSrvXact::acceptDataConnection() {
	Assert(theDataIoState == disListen);
	Assert(!theDataConn.sock());
	theDataConn.protoStat = protoStat;
//Comment << theSock.fd() << " accepting " << theDataConn.sock().fd() << endc;
	bool fatal = false;
	Must(theDataConn.accept(theSock, SockOpt(), fatal));
//Comment << theSock.fd() << " accepted " << theDataConn.sock().fd() << endc;

	stopDataListen();
	if (theDataXferState == dxsConnWaiting && !theDataConn.theWr.theReserv)
		theDataConn.theWr.start(this);
	theDataIoState = disBusy;
}

void FtpSrvXact::noteDataChannelWriteReady() {
	Assert(theDataIoState == disBusy);

	// resume writing RETR response and start data transfer if we were
	// waiting for the connection
	if (theDataXferState == dxsConnWaiting) {
		newState(stSpaceWaiting); // switch to writing ctrl response
		theConn->theRd.stop(this);
		if (!theConn->theWr.theReserv)
			theConn->theWr.start(this);
		if (const Error err = startDataXfer()) { // switch to dxsInProgress
			finish(err);
			return;
		}
	}

	Assert(theOid);
	Assert(theBodyIter);
	Assert(theDataXferState == dxsInProgress);

	theBodyIter->pour();
	const Size sz = theDataConn.write();

	if (sz > 0)
		theWrittenSize.got(sz);

	if ((theBodyIter->pouredAll() && theDataConn.theWrBuf.empty()) ||
		theDataConn.bad()) {
		theDataXferState = theDataConn.bad() ? dxsAborted : dxsComplete;
		closeDataChannel();
		theBodyIter->putBack();
		theBodyIter = 0;

		// tell the client we are done writing data
		if (!theConn->theWr.theReserv)
			theConn->theWr.start(this);
		if (theState == stHdrWaiting) // could be if we were in dxsConnWaiting
			newState(stSpaceWaiting);
	} else
	if (!theDataConn.theWr.theReserv)
		theDataConn.theWr.start(this);
}

Error FtpSrvXact::processRetr() {
	if (!theReq.argument())
		return errFtpNoArgument;
	theFilename = theReq.argument();

	if (theDataXferState != dxsNone)
		return errFtpCmdSequence;

	if (!(Should(theDataIoState != disNone))) // internal error?
		return errOther;

	if (theDataIoState == disListen) { // got RETR before data connection
		// wait for the data connection; TODO: add timeout
		theDataXferState = dxsConnWaiting;
		return 0;
	}

	return startDataXfer();
}

// called when we are ready to start writing response data
Error FtpSrvXact::startDataXfer() {
	if (!Should(theDataXferState == dxsNone ||
		theDataXferState == dxsConnWaiting)) // internal inconsistency?
		return errOther;

	if (!Should(theDataConn.sock())) // internal inconsistency?
		return errOther;

	if (!prepareContent()) {
		theDataXferState = dxsBadUrl;
		closeDataChannel(); // makeRetrResponse sends error
		return 0;
	}

	theBodyIter->start(&theDataConn.theWrBuf);
	theDataXferState = dxsInProgress;
	return 0;
}

bool FtpSrvXact::prepareContent() {
	if (theBodyIter)
		return true;

	const String url(theDir + '/' + theFilename);
	const char *p = url.data();
	theOid.parse(p, p + url.len());
	theOid.get(true);

	// XXX: pxy server cannot support ignoreUrls; see Server::hostIdx
	bool ignoreUrls = (TheSrvOpts.ignoreUrls && theOwner->hostIdx() >= 0);

	if (ignoreUrls)
		overwriteUrl();
	else
	if (theOid.foreignUrl())
		grokForeignUrl(url, ignoreUrls);

	if (theOid.foreignUrl())
		return false;
	else {
		Assert(theOid.type() >= TheContentMgr.normalContentStart());
		theRepContentCfg = TheContentMgr.get(theOid.type());
		theBodyIter = theRepContentCfg->getBodyIter(theOid);
		theWrittenSize.expectMore(theBodyIter->contentSize());
		return true;
	}
}

void FtpSrvXact::makeSimpleResponse(ostream &os, const String &reply) {
	os << reply;
}

void FtpSrvXact::makeMdtmResponse(ostream &os) {
	// XXX: hard coded modification time
	// should be similar to IMS
	os
		<< ftpRspMdtm
		<< "20080714151541"
		<< crlf
		;
}

void FtpSrvXact::makeSizeResponse(ostream &os) {
	if (prepareContent())
		os << ftpRspSize << (int)theBodyIter->contentSize() << crlf;
	else
		os << ftpRspBadUrl;
}

void FtpSrvXact::makePasvResponse(ostream &os) {
	Assert(theDataPort >= 0);
	os << ftpRspPasv;
	// XXX: IPv4 only
	for (int i = 0; i < 4; ++i)
		os << theOwner->host().octet(i) << ',';
	const int p1 = theDataPort / 256;
	const int p2 = theDataPort - 256 * p1;
	os << p1 << ',' << p2 << crlf;
	theDataPort = -1;
}

void FtpSrvXact::makeRetrResponse(ostream &os) {
	switch (theDataXferState) {
		case dxsNone:
			Should(false); // internal error
			break;
		case dxsConnWaiting:
			os << ftpRspRetrNoDataConn;
			break;
		case dxsInProgress:
			os << ftpRspRetrInProgress;
			break;
		case dxsBadUrl:
			Assert(theOid.foreignUrl());
			os << ftpRspBadUrl;
			theDataXferState = dxsNone;
			break;
		case dxsAborted: // should not happen because we respond first?
			os << ftpRspRetrAborted;
			theDataXferState = dxsNone;
			break;
		case dxsComplete: // should not happen because we respond first?
			os << ftpRspRetrComplete;
			theDataXferState = dxsNone;
			break;
	}
}
